#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include "scode.h"
#include "config.h"
#include "log.h"
#include "ep_init.h"
#include "ep_util.h"

const PCSTR Mod_UPIWIN = "upiwin";              /* name of the primary UPIWIN module */
const PCSTR Mod_UPIWIN_tmp = "upiwin_tmp";      /* name of the temporary UPIWIN module */

static wchar_t *python_name = NULL;       /* location of the Python executable */

PyObject *UPIWIN_module = NULL;
PyObject *UPIWIN_tmp_module = NULL;

static PyModuleDef DefUPIWIN_tmp = {
  PyModuleDef_HEAD_INIT,   /* standard garbage */
  Mod_UPIWIN,              /* module name */
  NULL,                    /* no doc string */
  -1,                      /* no per-module memory */
  NULL,                    /* method defs will be added later */
  NULL,                    /* no slots for multi-phase init */
  NULL,                    /* no traversal proc */
  NULL,                    /* no clear function */
  NULL                     /* no free function */
};

static PyObject *init_upiwin_tmp_module(void)
{
  PyObject *module = PyModule_Create(&DefUPIWIN_tmp);
  return module;
}

/* used to link the two modules into Python's init table */
static struct _inittab upiwin_inittab[] = {
  { Mod_UPIWIN,     Epython_init_upiwin_module },
  { Mod_UPIWIN_tmp, init_upiwin_tmp_module     },
  { NULL,           NULL                       }
};

static void epython_cleanup(void)
{
  Py_DECREF(UPIWIN_tmp_module);
  UPIWIN_tmp_module = NULL;
  Py_DECREF(UPIWIN_module);
  UPIWIN_module = NULL;
  if (!Py_FinalizeEx())
    Log(LWARN, "errors encountered when Python uninitialized itself");
  PyMem_RawFree(python_name);
  python_name = NULL;
}

HRESULT Epython_setup(void)
{
  HRESULT hr;
  size_t size;

  python_name = Py_DecodeLocale(Gconfig.python_loc, &size);
  if (!python_name)
  {
    if (size==(size_t)(-1))
    {
      Log(LFATAL, "error allocating Python program location");
      return E_OUTOFMEMORY;
    }
    else
    {
      Log(LFATAL, "internal error in Py_DecodeLocale");
      return E_UNEXPECTED;
    }
  }
  Py_SetProgramName(python_name);

  if (PyImport_ExtendInittab(upiwin_inittab))
  {
    Log(LFATAL, "error allocating extended init table");
    hr = E_OUTOFMEMORY;
    goto error_0;
  }

  Py_Initialize();

  /* Import the modules */
  UPIWIN_module = PyImport_ImportModule(Mod_UPIWIN);
  if (!UPIWIN_module)
  {
    Log(LFATAL, "error importing the upiwin module");
    hr = Epython_trace_exception();
    goto error_1;
  }
  UPIWIN_tmp_module = PyImport_ImportModule(Mod_UPIWIN_tmp);
  if (!UPIWIN_tmp_module)
  {
    Log(LFATAL, "error importing the upiwin_tmp module");
    hr = Epython_trace_exception();
    goto error_2;
  }

  hr = Config_exitfunc(epython_cleanup);
  if (FAILED(hr))
    epython_cleanup();
  return hr;

error_2:
  Py_DECREF(UPIWIN_module);
  UPIWIN_module = NULL;
error_1:
  Py_FinalizeEx();
error_0:
  PyMem_RawFree(python_name);
  python_name = NULL;
  return hr;
}
